/*
 * $Id: ngrep.h,v 1.13 2000/02/01 12:28:08 jpr5 Exp $
 *
 */

/* for PPPHDR_SIZE, stevens says 24, but on linux (at least) it's 4 */
/*   who knows if this will break anything */
#define ETHHDR_SIZE 14 
#define PPPHDR_SIZE 4 
#define SLIPHDR_SIZE 16
#define RAWHDR_SIZE 0
#define LOOPHDR_SIZE 4 
#define FDDIHDR_SIZE 21

#ifndef IP_OFFMASK
#define IP_OFFMASK 0x1fff
#endif

#define WORD_REGEX "((^%s\\W)|(\\W%s$)|(\\W%s\\W))"
#define IP_ONLY "ip and ( %s)"

char *get_filter(char **);
void process(u_char *, struct pcap_pkthdr*, u_char *);
int re_match_func(char *, int); 
int blank_match_func(char *, int); 
void dump(char *, int); 
void clean_exit(int);
void usage(int);
void version(void);
